CON

  { ==[ CLOCK SET ]== }       
  _CLKMODE      = XTAL1 + PLL16X
  _XINFREQ      = 5_000_000                             ' 5MHz Crystal

  MAX_INPUT_SIZE = 64                                   ' actual size of input string must be less than 32 bytes

OBJ

  DEBUG  : "FullDuplexSerial"    

VAR
                 
  LONG done
  LONG addresses[2]
  LONG w[80]
  BYTE nstr[MAX_INPUT_SIZE]
  BYTE ostr[20]
  LONG h[5]
  
  BYTE str[100]

PUB Main | tmp, retaddr, i

  DEBUG.start(31, 30, 0, 57600)
  waitcnt(clkfreq + cnt)  
  DEBUG.tx($D)   
  
  tmp := string("1235")
  bytemove(@str, tmp, strsize(tmp))
  'DEBUG.dec(strsize(tmp))
  'DEBUG.tx($D)

  retaddr := pasm_sha1(@str, strsize(@str)) 

  REPEAT i FROM 0 TO 19
    DEBUG.hex(byte[retaddr + i], 2)

  DEBUG.tx($D)


  repeat
    waitcnt(0)

PUB pasm_sha1 (inputAddr, length) | i

  bytefill(@nstr, 0, MAX_INPUT_SIZE)
  bytemove(@nstr, inputAddr, length)

  byte[@nstr][length] := $80 'append the bit '1' to the message
  byte[@nstr][63] := length << 3         ' CURRENT LIMIT IS 32 BYTES

  bytemove(@w, @nstr, 64)
                       
  addresses[0] := @ostr
  addresses[1] := @w

  'Extend the sixteen 32-bit words into eighty 32-bit words:
  REPEAT i FROM 16 TO 79
    w[i] := (w[i-3] ^ w[i-8] ^ w[i-14] ^ w[i-16]) <- 1


  done := false
  cognew(@entry, @done)

  REPEAT UNTIL (done)
  DEBUG.hex(done, 8)
  DEBUG.tx($D)
  RETURN @ostr
  
PUB sha1 (inputAddr, length) | i, a,b,c,d,e, f, k, temp

''Note 1: All variables are unsigned 32 bits and wrap modulo 232 when calculating
''Note 2: All constants in this pseudo code are in big endian. 
''        Within each word, the most significant byte is stored in the leftmost byte position
                
  bytefill(@nstr, 0, MAX_INPUT_SIZE)
  bytemove(@nstr, inputAddr, length)

  'Initialize variables:
  h[0] := $67452301
  h[1] := $EFCDAB89
  h[2] := $98BADCFE
  h[3] := $10325476
  h[4] := $C3D2E1F0
   
  'Pre-processing:                   
  {DEBUG.str(string("Length: "))
  DEBUG.dec(length)
  DEBUG.tx($D)   }
  byte[@nstr][length] := $80 'append the bit '1' to the message
  {DEBUG.str(string("Bytes to Add: ")) 
  DEBUG.dec((56 - ((length) // 64)))
  DEBUG.tx($D)  }
  byte[@nstr][63] := length << 3         ' CURRENT LIMIT IS 32 BYTES
'  DEBUG.dec(length << 3)
 ' DEBUG.tx($D)

  REPEAT i FROM 0 TO 63
    DEBUG.bin(byte[@nstr][i], 8)
    DEBUG.tx(" ")
    IF (i // 8 == 7)
      DEBUG.tx($D) 

    ' append 0 <= k < 512 bits '0', so that the resulting message length (in bits) is congruent to 448 == -64 (mod 512)
    ' append length of message (before pre-processing), in bits, as 64-bit big-endian integer
  {DEBUG.tx($D)
  DEBUG.dec(length + (56 - ((length + 1) // 64)) + 8)
  DEBUG.tx($D)
  DEBUG.dec(length << 3) }
  DEBUG.tx($D)
  {
  append 0 <= k < 512 bits '0', so that the resulting message length (in bits) is congruent to 448 == -64 (mod 512)
  append length of message (before pre-processing), in bits, as 64-bit big-endian integer
   
  'Process the message in successive 512-bit chunks:
  break message into 512-bit chunks
  for each chunk
      break chunk into sixteen 32-bit big-endian words w[i], 0 <= i <= 15
}
      bytemove(@w, @nstr, 64)

      'Extend the sixteen 32-bit words into eighty 32-bit words:
      REPEAT i FROM 16 TO 79
          w[i] := (w[i-3] ^ w[i-8] ^ w[i-14] ^ w[i-16]) <- 1

   
   
      'Initialize hash value for this chunk:
      a := h[0]
      b := h[1]
      c := h[2]
      d := h[3]
      e := h[4]
                     
      REPEAT i FROM 0 to 79
        CASE i
          0..19:             
            f := (b & c) | ((!b) & d)
            k := $5A827999
          20..39:
            f := b ^ c ^ d
            k := $6ED9EBA1
          40..59:
            f := (b & c) | (b & d) | (c & d) 
            k := $8F1BBCDC
          60..79:
            f := b ^ c ^ d
            k := $CA62C1D6

        temp := (a <- 5) + f + e + k + w[i]
        e := d
        d := c
        c := b <- 30
        b := a
        a := temp
   
      'Add this chunk's hash to result so far:
      h[0] := h[0] + a
      h[1] := h[1] + b 
      h[2] := h[2] + c
      h[3] := h[3] + d
      h[4] := h[4] + e
   
  'Produce the final hash value (big-endian):
  bytemove(@ostr, @h, 20)
  RETURN @ostr

DAT

                        ORG
entry
                        MOV     doneAddr, PAR
                        MOV     addr, doneAddr
                        ADD     addr, #4
                        RDLONG  ostrAddr, addr
                        ADD     addr, #4
                        RDLONG  wAddr, addr
                                                      
                        MOV     pa, h0                  ' a := h[0]
                        MOV     pb, h1                  ' b := h[1]
                        MOV     pc, h2                  ' c := h[2]
                        MOV     pd, h3                  ' d := h[3]
                        MOV     pe, h4                  ' e := h[4]

                        MOV     idx, #0                 ' REPEAT idx FROM 0 (to 79)
innerloop
                        CMP     idx, #19        WZ, WC  ' idx <= 19
              IF_A      JMP     #:sec2
                        MOV     pf, pb                  ' b
                        AND     pf, pc                  ' (b & c)
                        MOV     v2, pb                  ' b
                        XOR     v2, negone              ' (!b)
                        AND     v2, pd                  ' ((!b) & d)
                        OR      pf, v2                  ' f := (b & c) | ((!b) & d)
                        MOV     pk, consec1             ' k := $5A827999
                        
:sec2
                        CMP     idx, #20        WZ, WC  ' idx >= 20
              IF_B      JMP     #:sec3
                        CMP     idx, #39        WZ, WC  ' idx <= 39
              IF_A      JMP     #:sec3
                        MOV     pf, pb
                        XOR     pf, pc                  ' f := b ^ c 
                        XOR     pf, pd                  ' f := b ^ c ^ d
                        MOV     pk, consec2             ' k := $6ED9EBA1

:sec3
                        CMP     idx, #40        WZ, WC  ' idx >= 40
              IF_B      JMP     #:sec4
                        CMP     idx, #59        WZ, WC  ' idx <= 59
              IF_A      JMP     #:sec4
                        MOV     pf, pb                  ' b
                        AND     pf, pc                  ' (b & c)
                        MOV     v1, pb                  ' b
                        AND     v1, pd                  ' (b & d)
                        MOV     v2, pc                  ' c
                        AND     v2, pd                  ' (c & d)
                        OR      pf, v1                  ' f := (b & c) | (b & d)
                        OR      pf, v2                  ' f := (b & c) | (b & d) | (c & d)
                        MOV     pk, consec3             ' k := $8F1BBCDC 
            
:sec4
                        CMP     idx, #60        WZ, WC  ' idx >= 60
              IF_B      JMP     #:sec5
                        MOV     pf, pb                  ' b
                        XOR     pf, pc                  ' b ^ c
                        XOR     pf, pd                  ' f := b ^ c ^ d
                        MOV     pk, consec4             ' k := $CA62C1D6
            
:sec5
                        MOV     ptemp, pa               ' a
                        ROL     ptemp, #5               ' (a <- 5)
                        ADD     ptemp, pf               ' (a <- 5) + f
                        ADD     ptemp, pe               ' (a <- 5) + f + e
                        ADD     ptemp, pk               ' (a <- 5) + f + e + k

                        MOV     addr, wAddr             ' w
                        MOV     addrAdd, idx            ' i                       
                        SHL     addrAdd, #2             ' [i]
                        ADD     addr, addrAdd           ' w[i]
                        RDLONG  v1, addr                ' w[i]

                        ADD     ptemp, v1               ' temp := (a <- 5) + f + e + k + w[i]

                        MOV     pe, pd                  ' e := d
                        MOV     pd, pc                  ' d := c
                        MOV     v1, pb                  ' b
                        ROL     v1, #30                 ' b <- 30
                        MOV     pc, v1                  ' c := b <- 30
                        MOV     pb, pa                  ' b := a
                        MOV     pa, ptemp               ' a := temp

                        ADD     idx, #1                 ' idx++
                        CMP     idx, #80        WZ, WC
              IF_B      JMP     #innerloop


                        ADD     h0, pa                  ' h[0] := h[0] + a
                        ADD     h1, pb                  ' h[1] := h[1] + b
                        ADD     h2, pc                  ' h[2] := h[2] + c
                        ADD     h3, pd                  ' h[3] := h[3] + d
                        ADD     h4, pe                  ' h[4] := h[4] + e

                        MOV     addr, ostrAddr
                        WRLONG  h0, addr
                        ADD     addr, #4
                        WRLONG  h1, addr
                        ADD     addr, #4
                        WRLONG  h2, addr
                        ADD     addr, #4
                        WRLONG  h3, addr
                        ADD     addr, #4
                        WRLONG  h4, addr

                        WRLONG  negone, doneAddr        ' say done.

                        COGID   v1
                        COGSTOP v1

negone                  LONG    -1     

{consec1                 LONG    $5A827999     ' little endian
consec2                 LONG    $6ED9EBA1
consec3                 LONG    $8F1BBCDC
consec4                 LONG    $CA62C1D6 }
consec1                 LONG    $9979825A     ' big endian
consec2                 LONG    $A1EBD96E
consec3                 LONG    $DCBC1B8F
consec4                 LONG    $D6C162CA

h0                      RES
h1                      RES
h2                      RES
h3                      RES
h4                      RES

addr                    RES
v1                      RES
v2                      RES  
idx                     RES
addrAdd                 RES

doneAddr                RES 
ostrAddr                RES
wAddr                   RES

pa                      RES
pb                      RES
pc                      RES
pd                      RES
pe                      RES
pf                      RES
pk                      RES
ptemp                   RES

                        FIT
                        
PUB blah | i, a,b,c,d,e, f, k, temp 
   
      REPEAT i FROM 0 to 79
        CASE i
          0..19:             
            f := (b & c) | ((!b) & d)
            k := $5A827999
          20..39:
            f := b ^ c ^ d
            k := $6ED9EBA1
          40..59:
            f := (b & c) | (b & d) | (c & d) 
            k := $8F1BBCDC
          60..79:
            f := b ^ c ^ d
            k := $CA62C1D6 

        temp := (a <- 5) + f + e + k + w[i]
        e := d
        d := c
        c := b <- 30
        b := a
        a := temp
   
      'Add this chunk's hash to result so far:
      h[0] := h[0] + a
      h[1] := h[1] + b 
      h[2] := h[2] + c
      h[3] := h[3] + d
      h[4] := h[4] + e
   
  'Produce the final hash value (big-endian):
  bytemove(@ostr, @h, 20)
  RETURN @ostr

               